from __future__ import annotations
import typing

from tronpy.async_contract import AsyncContractMethod
from eth_typing import ChecksumAddress

from ...account import (
    Account,
    EthereumAccount,
    SolanaAccount,
    TronAccount,
)
from ...chain import Chain
from ...chain_client import (
    ChainClient,
    EthereumClient,
    SolanaClient,
    TronClient,
)
from ...transaction import (
    EthereumTransaction,
    SolanaTransaction,
    Transaction,
    TronTransaction,
)
from ... import config
from ..types import OrderData
from . import solana as solana_orders
from .solana.types import PlaceOrderParams
from .solana.instructions.place_order import PlaceOrderArgs

if typing.TYPE_CHECKING:
    from ... import MachClient


async def create_place_order_transaction[ChainType: Chain](
    *,
    client: MachClient,
    src_client: ChainClient[ChainType],
    order_data: OrderData,
    signer: Account[ChainType],
) -> Transaction[ChainType]:
    order_book_address = order_data.contract_address

    match src_client:
        case EthereumClient():
            contract = src_client.w3.eth.contract(
                address=typing.cast(ChecksumAddress, order_book_address),
                abi=config.ethereum_order_book_abi,
            )

            place_order_function = contract.functions.placeOrder(
                order_data.order_direction.to_eth(),
                order_data.order_funding.to_eth(),
                order_data.order_expiration.to_eth(),
                order_data.target_address,
                order_data.filler_address,
            )

            return await EthereumTransaction.from_contract_function(
                src_client,
                place_order_function,
                typing.cast(EthereumAccount, signer),
            )

        case SolanaClient():
            assert False, "TODO"
            place_order_accounts = solana_orders.get_accounts(
                client, src_client.chain, {}
            )

            place_order_instruction = solana_orders.place_order(
                args=PlaceOrderArgs(
                    params=PlaceOrderParams(
                        source_sell_amount=order_data.order_funding.src_amount_in,
                        min_sell_amount=0,
                        dest_token_mint=[],
                        dest_buy_amount=order_data.order_funding.dst_amount_out,
                        order_id=43,  # What is this?
                        # eid=chain_client.chain.layerzero_id,
                        eid=540,  # What is this?
                    )
                ),
                accounts=place_order_accounts,
                program_id=solana_orders.PROGRAM_ID,
            )

            return await SolanaTransaction.create(
                typing.cast(SolanaClient, src_client),
                (place_order_instruction,),
                (typing.cast(SolanaAccount, signer),),
            )

        case TronClient():
            contract = await src_client.native.get_contract(
                order_book_address
            )
            contract.abi = config.tron_order_book_abi

            return await TronTransaction.from_contract_method(
                src_client,
                typing.cast(TronAccount, signer),
                typing.cast(AsyncContractMethod, contract.functions.placeOrder),
                order_data.order_direction.to_tron(),
                order_data.order_funding.to_tron(),
                order_data.order_expiration.to_tron(),
                order_data.target_address,
                order_data.filler_address,
            )

        case _:
            raise NotImplementedError(src_client.chain)

__all__ = ["create_place_order_transaction"]


import abc
from abc import ABC
import dataclasses
from typing import Any


from ..chain import Chain


# Proxy for a chain-specific sent transaction
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class SentTransaction[ChainType: Chain](ABC):
    # TODO: This currently doesn't follow the convention in the rest of the codebase that the `.id` property should produce a CAIP identifier that is unique across chains. Instead we return the transaction hash or equivalent, which is not necessarily unique across chains.
    @property
    @abc.abstractmethod
    def id(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def chain(self) -> ChainType:
        pass

    # Confirms the transaction if necessary and returns the receipt/response
    @abc.abstractmethod
    async def wait_for_receipt(self, **kwargs) -> Any:
        pass


# Proxy for a chain-specific transaction
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class Transaction[
    ChainType: Chain,
](ABC):
    @abc.abstractmethod
    async def broadcast(self) -> SentTransaction[ChainType]:
        pass

    @property
    @abc.abstractmethod
    def chain(self) -> ChainType:
        pass



import abc
from abc import ABC

from aiohttp import ClientSession
from pydantic import TypeAdapter

from ... import config
from ...account import AccountBase
from ...chain import EthereumChain
from .. import utility
from .types import WalletPoints


class MilesServerMixin(ABC):
    __slots__ = ("__routes",)

    def __init__(self, url: str) -> None:
        # The double underscore forces name mangling to occur
        # This allows subclasses to have a different `self.routes` attribute
        self.__routes = config.config.miles_server.endpoints.add_url(url)

    @property
    @abc.abstractmethod
    def session(self) -> ClientSession:
        pass

    async def get_points(self, account: AccountBase) -> float:
        assert isinstance(
            account.chain, EthereumChain
        ), "Points only supported on Ethereum"

        url = f"{self.__routes.points}/{account.address}"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return WalletPoints.model_validate_json(bytes_result).points

    _wallet_points_validator = TypeAdapter(list[WalletPoints])

    async def get_all_points(self, limit: int = 10) -> list[WalletPoints]:
        params = {
            "limit": limit,
        }

        async with self.session.get(self.__routes.points, params=params) as response:
            bytes_result = await utility.to_bytes(response)

        return self._wallet_points_validator.validate_json(bytes_result)


__all__ = ["MilesServerMixin"]



from decimal import Decimal
from typing import Optional

from pydantic import BaseModel

from ..types import MachChain


class AssetInfo(BaseModel):
    id: str
    symbol: str
    name: str
    image: str
    current_price: Decimal
    market_cap: int
    market_cap_rank: int
    fully_diluted_valuation: int
    total_volume: int
    high_24h: Decimal
    low_24h: Decimal
    price_change_24h: Decimal
    price_change_percentage_24h: Decimal
    market_cap_change_24h: Decimal
    market_cap_change_percentage_24h: Decimal
    circulating_supply: Decimal
    total_supply: Decimal
    max_supply: Optional[Decimal]
    ath: Decimal
    ath_change_percentage: Decimal
    ath_date: str
    atl: Decimal
    atl_change_percentage: Decimal
    atl_date: str
    roi: Optional[Decimal]
    last_updated: str
    decimals: int


class AssetPricingData(BaseModel):
    chain: MachChain
    address: str
    symbol: str
    decimals: int
    price: float
    daily_percent_change: float


class UserAssetData(BaseModel):
    chain: MachChain
    address: str
    balance: int
    symbol: Optional[str] = None
    price: Optional[Decimal]
    daily_percent_change: Optional[Decimal]



from ..account import AccountBase
from ..asset import Token
from ..chain import Chain
from ..transaction import SentTransaction
from .scanner import Scanner


def address(chain: Chain, address: str) -> str:
    return Scanner.create(chain).address(address)


def account(account: AccountBase) -> str:
    return address(account.chain, account.address)


def transaction(chain: Chain, transaction_id: str) -> str:
    return Scanner.create(chain).transaction(transaction_id)


def sent_transaction(sent_transaction: SentTransaction) -> str:
    return transaction(sent_transaction.chain, sent_transaction.id)


def token(token: Token) -> str:
    return Scanner.create(token.chain).token(token)



from __future__ import annotations
import typing

from ..asset import Token
from .scanner import Scanner


class SolanaFM(Scanner):
    url = "https://solana.fm"

    @typing.override
    def address(self, address: str) -> str:
        return f"{self.url}/address/{address}"

    @typing.override
    def transaction(self, transaction_id: str) -> str:
        return f"{self.url}/tx/{transaction_id}"

    @typing.override
    def token(self, token: Token) -> str:
        return self.address(token.address)



from __future__ import annotations
import dataclasses
import typing
from typing import Optional

from eth_typing import ChecksumAddress
from web3.contract import AsyncContract
from web3.types import TxParams, TxReceipt

from .. import config
from ..account import EthereumAccount, EthereumAccountID
from ..chain import EthereumChain
from ..chain_client import EthereumClient
from ..transaction import EthereumSentTransaction, EthereumTransaction
from .token import ApprovableToken, NativeCoin


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class EthereumNativeCoin[Chain: EthereumChain](NativeCoin[Chain]):
    client: EthereumClient[Chain]

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain

    @typing.override
    async def get_balance(self, account_id: EthereumAccountID[Chain]) -> int:
        return await self.client.w3.eth.get_balance(account_id.address)

    @typing.override
    async def transfer(
        self,
        sender: EthereumAccount[Chain],
        recipient: EthereumAccountID[Chain],
        amount: int,
    ) -> tuple[EthereumSentTransaction[Chain], TxReceipt]:
        transaction = await EthereumTransaction.from_transaction_params(
            self.client,
            TxParams(
                {
                    "from": sender.native.address,
                    "to": recipient.address_,
                    "value": self.client.w3.to_wei(amount, "wei"),
                }
            ),
            sender,
        )

        sent_transaction = await transaction.broadcast()
        return sent_transaction, await sent_transaction.wait_for_receipt()


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class EthereumToken[Chain: EthereumChain](ApprovableToken[Chain]):
    client: EthereumClient[Chain]
    contract: AsyncContract

    @classmethod
    async def from_data(
        cls,
        client: EthereumClient[Chain],
        address: str,
        symbol: Optional[str],
        decimals: Optional[int],
    ) -> EthereumToken[Chain]:
        contract = client.w3.eth.contract(
            address=client.w3.to_checksum_address(address),
            abi=config.erc20_abi,
        )

        if not symbol:
            symbol = typing.cast(str, await contract.functions.symbol().call())

        if not decimals:
            decimals = typing.cast(int, await contract.functions.decimals().call())

        return cls(client=client, contract=contract, symbol=symbol, decimals=decimals)

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain

    # https://namespaces.chainagnostic.org/eip155/caip19
    @property
    @typing.override
    def asset_namespace(self) -> str:
        return "erc20"

    @property
    @typing.override
    def address(self) -> ChecksumAddress:
        return self.contract.address

    @typing.override
    async def get_balance(self, account_id: EthereumAccountID[Chain]) -> int:
        return await self.contract.functions.balanceOf(account_id.address).call()

    async def create_transfer_transaction(
        self,
        sender: EthereumAccount[Chain],
        recipient: EthereumAccountID[Chain],
        amount: int,
    ) -> EthereumTransaction[Chain]:
        contract_function = self.contract.functions.transfer(recipient.address, amount)
        return await EthereumTransaction.from_contract_function(
            self.client, contract_function, sender
        )

    @typing.override
    async def transfer(
        self,
        sender: EthereumAccount[Chain],
        recipient: EthereumAccountID[Chain],
        amount: int,
    ) -> tuple[EthereumSentTransaction[Chain], TxReceipt]:
        transaction = await self.create_transfer_transaction(sender, recipient, amount)
        sent_transaction = await transaction.broadcast()
        return sent_transaction, await sent_transaction.wait_for_receipt()

    @typing.override
    async def get_allowance(
        self,
        owner: EthereumAccountID[Chain],
        spender: EthereumAccountID[Chain],
    ) -> int:
        return await self.contract.functions.allowance(
            owner.address, spender.address
        ).call()

    async def create_approval_transaction(
        self,
        owner: EthereumAccount[Chain],
        spender: EthereumAccountID[Chain],
        amount: int,
    ) -> EthereumTransaction[Chain]:
        contract_function = self.contract.functions.approve(spender.address, amount)
        return await EthereumTransaction.from_contract_function(
            self.client, contract_function, owner
        )

    @typing.override
    async def approve(
        self,
        owner: EthereumAccount[Chain],
        spender: EthereumAccountID[Chain],
        amount: int,
    ) -> TxReceipt:
        transaction = await self.create_approval_transaction(owner, spender, amount)
        sent_transaction = await transaction.broadcast()
        return await sent_transaction.wait_for_receipt()



from __future__ import annotations
import typing

from eth_typing import ChecksumAddress
from solders.keypair import Keypair
from solders.pubkey import Pubkey

from ... import config
from ...account import (
    Account,
    AccountID,
    EthereumAccount,
    EthereumAccountID,
    SolanaAccount,
    SolanaAccountID,
)
from ...asset import SolanaToken, Token
from ...chain import Chain
from ...chain_client import (
    ChainClient,
    EthereumClient,
    SolanaClient,
)
from ...transaction import (
    EthereumTransaction,
    SolanaTransaction,
    Transaction,
)
from ...utility import solana as solana_utility
from ..types import OrderData
from .solana import instructions
from .solana.instructions import (
    DepositForBurnAccounts,
    DepositForBurnArgs,
)
from .solana.types import DepositForBurnParams

if typing.TYPE_CHECKING:
    from ...client import MachClient


async def create_cctp_burn_transaction[SrcChainType: Chain, DestChainType: Chain](
    *,
    client: MachClient,
    src_client: ChainClient[SrcChainType],
    src_token: Token[SrcChainType],
    dest_token: Token[DestChainType],
    order_data: OrderData,
    signer: Account[SrcChainType],
    recipient: AccountID[DestChainType],
) -> Transaction[SrcChainType]:
    token_messenger_minter_address = client.cctp_token_messenger_minter_address(
        src_client.chain
    )
    destination_domain = client.cctp_domain(recipient.chain)

    match src_client:
        case EthereumClient():
            recipient_bytes = recipient.encode_address().rjust(32, b"\0")

            contract = src_client.w3.eth.contract(
                address=typing.cast(ChecksumAddress, token_messenger_minter_address),
                abi=config.ethereum_cctp_token_messenger_abi,
            )

            burn_function = contract.functions.depositForBurn(
                order_data.order_funding.src_amount_in,
                destination_domain,
                recipient_bytes,
                order_data.order_direction.src_token_address,
            )

            return await EthereumTransaction.from_contract_function(
                src_client, burn_function, typing.cast(EthereumAccount, signer)
            )

        case SolanaClient():
            recipient_pubkey = typing.cast(SolanaAccountID, recipient).pubkey
            recipient_token_account = typing.cast(
                SolanaToken, dest_token
            ).associated_token_account(recipient_pubkey)

            params = DepositForBurnParams(
                amount=order_data.order_funding.src_amount_in,
                destination_domain=destination_domain,
                mint_recipient=recipient_token_account,
            )

            args = DepositForBurnArgs(params=params)

            owner = typing.cast(SolanaAccountID, signer).pubkey
            burn_token_account = typing.cast(
                SolanaToken, src_token
            ).associated_token_account(owner)

            message_transmitter = Pubkey.from_string(
                client.cctp_message_transmitter_address(src_client.chain)
            )
            token_messenger_minter = Pubkey.from_string(token_messenger_minter_address)

            usdc_mint = typing.cast(SolanaToken, src_token).mint

            accounts = DepositForBurnAccounts(
                owner=owner,
                event_rent_payer=owner,
                sender_authority_pda=solana_utility.find_program_address(
                    "sender_authority", token_messenger_minter
                )[0],
                burn_token_account=burn_token_account,
                message_transmitter=solana_utility.find_program_address(
                    "message_transmitter", message_transmitter
                )[0],
                token_messenger=solana_utility.find_program_address(
                    "token_messenger", token_messenger_minter
                )[0],
                remote_token_messenger=solana_utility.find_program_address(
                    "remote_token_messenger",
                    token_messenger_minter,
                    (str(destination_domain),),
                )[0],
                token_minter=solana_utility.find_program_address(
                    "token_minter", token_messenger_minter
                )[0],
                local_token=solana_utility.find_program_address(
                    "local_token",
                    token_messenger_minter,
                    (usdc_mint,),
                )[0],
                burn_token_mint=usdc_mint,
                message_sent_event_data=Keypair().pubkey(),
                message_transmitter_program=message_transmitter,
                token_messenger_minter_program=token_messenger_minter,
                event_authority=solana_utility.find_program_address(
                    "__event_authority", token_messenger_minter
                )[0],
                program=token_messenger_minter,
            )

            instruction = instructions.deposit_for_burn(args=args, accounts=accounts)

            return await SolanaTransaction.create(
                src_client, (instruction,), (typing.cast(SolanaAccount, signer),)
            )

        case _:
            raise NotImplementedError(src_client.chain)


__all__ = ["create_cctp_burn_transaction"]



from __future__ import annotations
import asyncio
from decimal import Decimal
import itertools
import logging
from typing import Any, Optional

from aiohttp import ClientSession
import cachebox
from cachebox import Cache
from pydantic import TypeAdapter

from ... import config
from ...account import AccountBase, AccountIDManager
from ...asset import Token
from ...chain import Chain, SupportedChain
from ...chain_client import ChainClient
from ...log import LogContextAdapter, Logger
from .. import utility
from ..types import MachChain
from .types import AssetInfo, AssetPricingData, UserAssetData


# This is a completely separate client instead of a mixin like the miles server because it has to be initialized, which is expensive and unnecessary for some use cases
class AssetServer:
    __slots__ = (
        "logger",
        "routes",
        "tokens",
        "session",
    )

    @classmethod
    @cachebox.cached(Cache(0), copy_level=2)
    async def create(
        cls,
        *,
        logger: Logger = logging.getLogger("mach-client"),
    ) -> AssetServer:
        client = cls(logger)
        await client.refresh_assets()
        return client

    def __init__(self, logger: Logger):
        self.logger = LogContextAdapter(logger, "Asset Server")
        self.routes = config.config.token_server.endpoints.add_url(
            config.config.token_server.url
        )
        self.tokens: dict[Chain, set[Token]] = {}
        self.session = ClientSession()
        self.session.headers.update(
            (
                ("accept", "application/json"),
                ("Content-Type", "application/json"),
            )
        )

    async def close(self) -> None:
        await self.session.close()

    def is_supported(self, token: Token) -> bool:
        return token.chain in self.tokens and token in self.tokens[token.chain]

    async def _process_chain_data(
        self, chain: Chain, chain_data: dict[str, dict[str, Any]]
    ) -> None:
        chain_client = await ChainClient.create(chain)

        if chain not in self.tokens:
            self.tokens[chain] = set()

        self.tokens[chain].update(
            await asyncio.gather(
                *[
                    Token.register(
                        client=chain_client,
                        address=address,
                        # TODO: The endpoint returns lowercased symbols
                        symbol=token_data["symbol"].upper(),
                        decimals=token_data["decimals"],
                    )
                    for address, token_data in chain_data.items()
                ]
            )
        )

    _assets_validator = TypeAdapter(dict[MachChain, dict[str, dict[str, Any]]])

    # This registers all the tokens returned by the GET /assets endpoint with the client
    async def refresh_assets(self) -> None:
        async with self.session.get(self.routes.assets) as response:
            bytes_result = await utility.to_bytes(response)

        await asyncio.gather(
            *[
                self._process_chain_data(chain.to_chain(), chain_data)
                for chain, chain_data in self._assets_validator.validate_json(
                    bytes_result
                ).items()
            ]
        )

    async def _get_asset_info(self, chain: Chain, address: str) -> Optional[AssetInfo]:
        mach_chain = MachChain.from_chain(chain)
        url = f"{self.routes.assets}/{mach_chain.value}/{address}"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        if bytes_result == b"null":
            return None

        return AssetInfo.model_validate_json(bytes_result)

    # Call this if the token in question hasn't yet been registered
    # This gets the asset info and registers the token
    async def get_asset_info(
        self, chain: Chain, address: str
    ) -> Optional[tuple[Token, AssetInfo]]:
        if not (asset_info := await self._get_asset_info(chain, address)):
            return None

        # The refresh_assets method won't register every possible token that you could get info for (ie. it omits very obscure ones)
        # So it's possible we've never seen this token before, and thus should register it first so that it can be looked up
        token = await Token.register(
            client=await ChainClient.create(chain),
            address=address,
            symbol=asset_info.symbol.upper(),
            decimals=asset_info.decimals,
        )

        return (token, asset_info)

    # Call this if the token in question is already in the `Token.lookup_cache`
    async def get_token_info(self, token: Token) -> AssetInfo:
        # The logic is that if you were able to construct a Token object, then it must be in the lookup cache
        # If it was in the lookup cache (which was populated by the asset server), then it must have an asset info
        return await self._get_asset_info(token.chain, token.address)  # type: ignore

    async def get_pricing_data(self, token: Token) -> AssetPricingData:
        mach_chain = MachChain.from_chain(token.chain)
        url = f"{self.routes.prices}/{mach_chain.value}/{token.address}"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return AssetPricingData.model_validate_json(bytes_result)

    async def get_price(self, token: Token) -> Decimal:
        pricing_data = await self.get_pricing_data(token)
        return Decimal(pricing_data.price)

    _asset_data_validator = TypeAdapter(dict[MachChain, list[UserAssetData]])

    async def get_raw_token_balances(
        self, account: AccountBase
    ) -> dict[MachChain, list[UserAssetData]]:
        url = f"{self.routes.users}/{account.address}/assets"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return self._asset_data_validator.validate_json(bytes_result)

    _chain_asset_data_validator = TypeAdapter(list[UserAssetData])

    async def get_raw_chain_token_balances(
        self, account: AccountBase
    ) -> list[UserAssetData]:
        mach_chain = MachChain.from_chain(account.chain)
        url = f"{self.routes.users}/{account.address}/assets/{mach_chain.value}"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return self._chain_asset_data_validator.validate_json(bytes_result)

    async def _process_chain_balance_data(
        self, chain: Chain, balance_data: list[UserAssetData]
    ) -> dict[Token, int]:
        chain_client = await ChainClient.create(chain)

        tokens = await asyncio.gather(
            *[
                Token.register(
                    client=chain_client,
                    address=asset_data.address,
                    symbol=asset_data.symbol,
                    decimals=None,
                )
                for asset_data in balance_data
                if asset_data.address != "native"
            ]
        )

        balances = dict(zip(tokens, map(lambda data: data.balance, balance_data)))
        return balances

    async def get_token_balances(
        self, account: AccountBase
    ) -> dict[Chain, dict[Token, int]]:
        raw_balances = await self.get_raw_token_balances(account)

        account_chain = type(account.chain)

        filtered_balances = [
            (chain, balances)
            for chain, balances in map(
                lambda item: (item[0].to_chain(), item[1]),
                raw_balances.items(),
            )
            if isinstance(chain, account_chain)
        ]

        chains = map(lambda item: item[0], filtered_balances)

        result = await asyncio.gather(
            *[
                self._process_chain_balance_data(chain, balance_data)
                for chain, balance_data in filtered_balances
            ]
        )

        return dict(zip(chains, result))

    async def get_token_balances_in_coins(
        self, account_id: AccountBase
    ) -> dict[Chain, dict[Token, Decimal]]:
        raw_balances = await self.get_token_balances(account_id)
        return utility.balances_in_coins(raw_balances)

    async def get_chain_token_balances(self, account: AccountBase) -> dict[Token, int]:
        raw_balances = await self.get_raw_chain_token_balances(account)
        return await self._process_chain_balance_data(account.chain, raw_balances)

    async def get_chain_token_balances_in_coins(
        self, account: AccountBase
    ) -> dict[Token, Decimal]:
        raw_balances = await self.get_chain_token_balances(account)
        return utility.chain_balances_in_coins(raw_balances)

    async def get_all_raw_token_balances(
        self, accounts: AccountIDManager
    ) -> dict[Chain, list[UserAssetData]]:
        all_accounts = [
            accounts[chain]
            for chain in (
                SupportedChain.ETHEREUM.value,
                SupportedChain.SOLANA.value,
                SupportedChain.TRON.value,
            )
            if chain in accounts
        ]

        balances = await asyncio.gather(
            *[self.get_raw_token_balances(account) for account in all_accounts]
        )

        result: dict[Chain, list[UserAssetData]] = {}

        for account, balances in zip(all_accounts, balances):
            result.update(
                (
                    (chain.to_chain(), chain_balance)
                    for chain, chain_balance in balances.items()
                    if isinstance(chain.to_chain(), type(account.chain))
                )
            )

        return result

    async def get_all_token_balances(
        self, accounts: AccountIDManager
    ) -> dict[Chain, dict[Token, int]]:
        balances = await asyncio.gather(
            *[
                self.get_token_balances(accounts[chain])
                for chain in (
                    SupportedChain.ETHEREUM.value,
                    SupportedChain.SOLANA.value,
                    SupportedChain.TRON.value,
                )
                if chain in accounts
            ]
        )

        return dict(
            itertools.chain(*[chain_balances.items() for chain_balances in balances])
        )

    async def get_all_token_balances_in_coins(
        self, accounts: AccountIDManager
    ) -> dict[Chain, dict[Token, Decimal]]:
        balances = await self.get_all_token_balances(accounts)
        return utility.balances_in_coins(balances)


__all__ = ["AssetServer"]
